module Auto(ToLaTeX(toLaTeX),Auto(),newAuto,(-/),(-//),(/->),(<:),accept,Node(),node,startNode,startTerminalNode,terminalNode ) where
import Data.List

class ToLaTeX a where
  toLaTeX :: a -> String

data Node name = Node name Bool Bool

instance Eq n => Eq (Node n) where
  (Node n1 _ _) == (Node n2 _ _) = n1==n2


instance Ord n => Ord (Node n) where
  compare (Node n1 _ _)  (Node n2 _ _) = compare n1 n2


ntoLaTeX neighbour  (Node name False False) =
  "\\node [state] ("++fs name++") "++neighbour++" {$"++fs name++"$};\n"
ntoLaTeX neighbour (Node name True False) =
  "\\node [state, initial] ("++fs name++") "++neighbour++" {$"++fs name++"$};\n"
ntoLaTeX neighbour (Node name False True) =
  "\\node [state, accepting] ("++fs name++") "++neighbour++" {$"++fs name++"$};\n"
ntoLaTeX neighbour(Node name _ _) =
    "\\node [state, initial, accepting] ("++fs name++") "++neighbour++" {$"++fs name++"$};\n"

fs xs = (filter (\x->not $elem x "\"[]" ))$show xs
 
node n = Node n False False   

data Trans name symbol = Trans (Node name) [symbol] (Node name) 

instance (Ord n,Show n, Show s) =>  ToLaTeX (Trans n s) where
  toLaTeX (Trans (Node start _ _) sym (Node end _ _)) 
    |start==end ="("++fs start++") edge [loop ,above] node {"++fs sym++"} ("++fs end++")\n"
    |start > end = "("++fs start++") edge [bend left, below] node {"++fs sym++"} ("++fs end++")\n"
    |otherwise = "("++fs start++") edge [above] node {"++fs sym++"} ("++fs end++")\n"

data Auto name symbol = Auto [Trans name symbol]

instance (Show n, Ord n,Show s) =>  ToLaTeX (Auto n s) where
  toLaTeX auto@(Auto trs) =
       "\\begin{tikzpicture}[->,>=stealth',semithick ,node  distance =2cm]\n"
    ++ (nodesToLaTeX "" $nodes auto)
    ++ "\n\\path "
    ++ concat (map toLaTeX trs)
    ++ ";"
    ++ "\n\\end{tikzpicture}"
   where
    nodesToLaTeX neigh [] = ""
    nodesToLaTeX neigh [x] = ntoLaTeX neigh x
    nodesToLaTeX neigh (x@(Node n _ _):xs) =
      ntoLaTeX neigh x ++ (nodesToLaTeX ("[right of="++fs n++"]") xs)

infixl 2 <: 
infix 6 -/
infix 6 -//   
infix 4 /-> 

(-/) node symbol =  Trans node [symbol]
(-//) node symbols =  Trans node symbols

(/->) start node2 = start node2 

(<:) (Auto trs) tr@(Trans s1 s e1) = Auto ((combine isit):isitnot)
  where
    (isit,isitnot) = partition (\(Trans s2 _ e2) -> s1==s2 && e1==e2) trs
    combine =
      foldl (\(Trans s sym1 e) (Trans _ sym2 _)->(Trans s1 (sym1++sym2) e)) tr

nodes (Auto trs) = sort$nub$ concat [[n1,n2]|(Trans n1 _ n2)<-trs]

data State name symbol = State name (Auto name symbol) 

step (State name (Auto trs)) (x:xs)
  = [n2|(Trans  (Node n _ _) syms (Node n2 _ _))<-trs,n==name,elem x syms]

accept auto@(Auto trs) xs = reduce  (State start auto) xs
  where
    [start] = nub [n|(Trans  (Node n True _) s _)<-trs]
    isTerminal (Node _ _ True) = True
    isTerminal _ = False
    nodeName (Node n _ _) = n
    reduce (State n1 (Auto trs)) []
     = not$null$filter (\n->isTerminal n&& nodeName n == n1)$nodes auto
    reduce st@(State n1 auto ) xs
      |null n2s = False
      |otherwise = reduce (State (head n2s) auto) $tail xs
     where 
       n2s = step st xs
     
      
newAuto :: Auto String Char
newAuto = Auto []

startNode x = Node x True False
terminalNode x = Node x False True
startTerminalNode x = Node x True True
